import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestOtherCommands:

    def test_commands_not_blocked(self):
        """ there is no reason to really block commands and operations over editable packages
        except for others doing an install that depends on the editable
        """
        t = TestClient(default_server_user=True)
        t.save({'conanfile.py': GenConanfile("lib", "0.1"),
                "test_package/conanfile.py": GenConanfile().with_test("pass")})
        t.run('editable add .')

        # Nothing in the cache
        t.run("list *")
        assert "There are no matching recipe references" in t.out
        t.run('list lib/0.1:*')
        assert "ERROR: Recipe 'lib/0.1' not found" in t.out

        t.run('export . ')
        assert "lib/0.1: Exported" in t.out
        t.run("list *")
        assert "lib/0.1" in t.out
        t.run('list lib/0.1:*')
        assert "PID:" not in t.out  # One binary is listed

        t.run('export-pkg .')
        assert "lib/0.1: Exporting package" in t.out

        t.run('list lib/0.1:*')
        assert "lib/0.1" in t.out  # One binary is listed

        t.run('upload lib/0.1 -r default')
        assert "Uploading recipe 'lib/0.1" in t.out

        t.run("remove * -c")
        # Nothing in the cache
        t.run("list *")
        assert "There are no matching recipe references" in t.out
        t.run('list lib/0.1:*')
        assert "ERROR: Recipe 'lib/0.1' not found" in t.out

    def test_create_editable(self):
        """
        test that an editable can be built with conan create
        """
        t = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           class Pkg(ConanFile):
               name = "lib"
               version = "0.1"
               def build(self):
                   self.output.info("MYBUILDFOLDER: {}".format(self.build_folder))
               """)
        t.save({'conanfile.py': conanfile,
                "test_package/conanfile.py": GenConanfile().with_test("pass"),
                "consumer/conanfile.txt": "[requires]\nlib/0.1"})
        t.run('editable add .')

        t.run("list *")
        assert "There are no matching" in t.out

        t.run("create .")
        package_id = "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        t.assert_listed_require({"lib/0.1": "Editable"})
        t.assert_listed_binary({"lib/0.1": (package_id,
                                            "EditableBuild")})
        assert f"lib/0.1: MYBUILDFOLDER: {t.current_folder}" in t.out
        t.run("list *")
        assert "lib/0.1" in t.out  # Because the create actually exports, TODO: avoid exporting?

        t.run("install consumer --build=*")
        t.assert_listed_require({"lib/0.1": "Editable"})
        t.assert_listed_binary({"lib/0.1": (package_id,
                                            "EditableBuild")})
        assert f"lib/0.1: MYBUILDFOLDER: {t.current_folder}" in t.out

        t.run("install consumer --build=editable")
        t.assert_listed_require({"lib/0.1": "Editable"})
        t.assert_listed_binary({"lib/0.1": (package_id,
                                            "EditableBuild")})
        assert f"lib/0.1: MYBUILDFOLDER: {t.current_folder}" in t.out



import pytest
from parameterized import parameterized

from conan.internal.graph.graph_error import GraphMissingError, GraphLoopError, GraphConflictError
from conan.errors import ConanException
from test.integration.graph.core.graph_manager_base import GraphManagerTest
from conan.test.utils.tools import GenConanfile


def _check_transitive(node, transitive_deps):
    values = list(node.transitive_deps.values())

    assert len(values) == len(transitive_deps), f"{node}: Number of deps don't match " \
                                                f"\n{[r.require.ref for r in values]}" \
                                                f"!=\n{transitive_deps}"

    for v1, v2 in zip(values, transitive_deps):
        assert v1.node is v2[0], f"{v1.node}!={v2[0]}"
        assert v1.require.headers is v2[1], f"{v1.node}!={v2[0]} headers"
        assert v1.require.libs is v2[2], f"{v1.node}!={v2[0]} libs"
        assert v1.require.build is v2[3], f"{v1.node}!={v2[0]} build"
        assert v1.require.run is  v2[4], f"{v1.node}!={v2[0]} run"
        assert len(v2) <= 5


class TestLinear(GraphManagerTest):
    def test_basic(self):
        deps_graph = self.build_graph(GenConanfile("app", "0.1"))
        assert 1 == len(deps_graph.nodes)
        node = deps_graph.root
        self._check_node(node, "app/0.1")

    def test_dependency(self):
        # app -> libb0.1
        self.recipe_cache("libb/0.1")
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])
        deps_graph = self.build_consumer(consumer)

        assert 2 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[], dependents=[app])

    def test_dependency_missing(self):
        # app -> libb0.1 (non existing)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])
        deps_graph = self.build_consumer(consumer, install=False)

        # TODO: Better error handling
        assert type(deps_graph.error) == GraphMissingError

        assert 1 == len(deps_graph.nodes)
        app = deps_graph.root
        self._check_node(app, "app/0.1", deps=[])

    def test_transitive(self):
        # app -> libb0.1 -> liba0.1
        # By default if packages do not specify anything link=True is propagated run=None (unknown)
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, True, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_transitive_propagate_link(self):
        # app -> libb0.1 -> liba0.1
        # transitive_link=False will avoid propagating linkage requirement
        self.recipe_cache("liba/0.1")
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1",
                                                                          transitive_libs=False))
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, False, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_transitive_all_static(self):
        # app -> libb0.1 (static) -> liba0.1 (static)
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=False)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        for r, t in libb.transitive_deps.items():
            assert r.package_id_mode == "minor_mode"

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, False, True, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_transitive_all_static_transitive_headers(self):
        # app -> libb0.1 (static) -> liba0.1 (static)
        self.recipe_cache("liba/0.1", option_shared=False)
        libb = GenConanfile().with_requirement("liba/0.1", transitive_headers=True)
        libb.with_shared_option()
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, True, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_transitive_all_shared(self):
        # app -> libb0.1 (shared)  -> liba0.1 (shared)
        self.recipe_cache("liba/0.1", option_shared=True)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=True)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        # Default for app->liba is that it doesn't link, libb shared will isolate symbols by default
        _check_transitive(app, [(libb, True, True, False, True),
                                (liba, False, False, False, True)])
        _check_transitive(libb, [(liba, True, True, False, True)])

    def test_transitive_all_shared_transitive_headers_libs(self):
        # app -> libb0.1 (shared) -> liba0.1 (shared)
        self.recipe_cache("liba/0.1", option_shared=True)
        libb = GenConanfile().with_requirement("liba/0.1", transitive_headers=True,
                                               transitive_libs=True)
        libb.with_shared_option(True)
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        # Default for app->liba is that it doesn't link, libb shared will isolate symbols by default
        _check_transitive(app, [(libb, True, True, False, True),
                                (liba, True, True, False, True)])
        _check_transitive(libb, [(liba, True, True, False, True)])

    def test_middle_shared_up_static(self):
        # app -> libb0.1 (shared) -> liba0.1 (static)
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=True)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, True),
                                (liba, False, False, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_middle_shared_up_static_transitive_headers(self):
        # app -> libb0.1 (shared) -> liba0.1 (static)
        self.recipe_cache("liba/0.1", option_shared=False)
        libb = GenConanfile().with_requirement("liba/0.1", transitive_headers=True)
        libb.with_shared_option(True)
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, True),
                                (liba, True, False, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_middle_static_up_shared(self):
        # app -> libb0.1 (static) -> liba0.1 (shared)
        self.recipe_cache("liba/0.1", option_shared=True)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=False)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, False, True, False, True)])
        _check_transitive(libb, [(liba, True, True, False, True)])

    def test_middle_static_up_shared_transitive_headers(self):
        # app -> libb0.1 (static) -> liba0.1 (shared)
        self.recipe_cache("liba/0.1", option_shared=True)
        libb = GenConanfile().with_requirement("liba/0.1", transitive_headers=True)
        libb.with_shared_option(False)
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, True, False, True)])
        _check_transitive(libb, [(liba, True, True, False, True)])

    def test_private(self):
        # app -> libb0.1 -(private) -> liba0.1
        self.recipe_cache("liba/0.1")
        libb = GenConanfile().with_requirement("liba/0.1", visible=False)
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_generic_library_without_shared_option(self):
        # app -> libb0.1 -> liba0.1 (library without shared option)
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("library"))
        libb = GenConanfile().with_requirement("liba/0.1")
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        with pytest.raises(ConanException) as exc:
            self.build_consumer(consumer)
        assert "liba/0.1: Package type is 'library', but no 'shared' option declared" in str(exc)

    def test_build_script_requirement(self):
        # app -> libb0.1 -br-> liba0.1 (build-scripts)
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("build-scripts"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_tool_requirement("liba/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False)])
        _check_transitive(libb, [(liba, False, False, True, True)])

    @parameterized.expand([("application",), ("shared-library",), ("static-library",),
                           ("header-library",), ("build-scripts",), (None,)])
    def test_generic_build_require_adjust_run_with_package_type(self, package_type):
        # app --br-> cmake (app)
        self.recipe_conanfile("cmake/0.1", GenConanfile().with_package_type(package_type))
        # build require with run=None by default
        consumer = self.recipe_consumer("app/0.1", build_requires=["cmake/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 2 == len(deps_graph.nodes)
        app = deps_graph.root
        cmake = app.edges[0].dst

        # node, headers, lib, build, run
        run = package_type in ("application", "shared-library", "build-scripts")
        _check_transitive(app, [(cmake, False, False, True, run)])

    def test_direct_header_only(self):
        # app -> liba0.1 (header_only)
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["liba/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 2 == len(deps_graph.nodes)
        app = deps_graph.root
        liba = app.edges[0].dst

        self._check_node(app, "app/0.1", deps=[liba])
        self._check_node(liba, "liba/0.1#123", dependents=[app])

        # node, headers, lib, build, run
        _check_transitive(app, [(liba, True, False, False, False)])

    def test_header_only(self):
        # app -> libb0.1 -> liba0.1 (header_only)
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library"))
        libb = GenConanfile().with_requirement("liba/0.1")
        self.recipe_conanfile("libb/0.1", libb)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, False, False, False, False)])
        _check_transitive(libb, [(liba, True, False, False, False)])

    def test_header_only_with_transitives(self):
        # app -> liba0.1(header) -> libb0.1 (static)
        #             \-----------> libc0.1 (shared)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("shared-library"))
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library")
                                                        .with_requires("libb/0.1", "libc/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["liba/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        liba = app.edges[0].dst
        libb = liba.edges[0].dst
        libc = liba.edges[1].dst

        self._check_node(app, "app/0.1", deps=[liba])
        self._check_node(liba, "liba/0.1#123", deps=[libb, libc], dependents=[app])
        self._check_node(libb, "libb/0.1#123", dependents=[liba])
        self._check_node(libc, "libc/0.1#123", dependents=[liba])

        # node, headers, lib, build, run
        _check_transitive(app, [(liba, True, False, False, False),
                                (libb, True, True, False, False),
                                (libc, True, True, False, True)])
        _check_transitive(liba, [(libb, True, True, False, False),
                                 (libc, True, True, False, True)])

    def test_multiple_header_only_with_transitives(self):
        # app -> libd0.1(header) -> liba0.1(header) -> libb0.1 (static)
        #                               \-----------> libc0.1 (shared)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("shared-library"))
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library")
                              .with_requires("libb/0.1", "libc/0.1"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_package_type("header-library")
                              .with_requires("liba/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        liba = libd.edges[0].dst
        libb = liba.edges[0].dst
        libc = liba.edges[1].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", deps=[libb, libc], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", dependents=[liba])
        self._check_node(libc, "libc/0.1#123", dependents=[liba])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (liba, True, False, False, False),
                                (libb, True, True, False, False),
                                (libc, True, True, False, True)])
        _check_transitive(libd, [(liba, True, False, False, False),
                                 (libb, True, True, False, False),
                                 (libc, True, True, False, True)])
        _check_transitive(liba, [(libb, True, True, False, False),
                                 (libc, True, True, False, True)])

    def test_static_multiple_header_only_with_transitives(self):
        # app -> libd0.1(static) -> liba0.1(header) -> libb0.1 (static)
        #                               \-----------> libc0.1 (shared)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("shared-library"))
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library")
                              .with_requires("libb/0.1", "libc/0.1"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_package_type("static-library")
                              .with_requires("liba/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        liba = libd.edges[0].dst
        libb = liba.edges[0].dst
        libc = liba.edges[1].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", deps=[libb, libc], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", dependents=[liba])
        self._check_node(libc, "libc/0.1#123", dependents=[liba])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False),
                                (liba, False, False, False, False),
                                (libb, False, True, False, False),
                                (libc, False, True, False, True)])
        _check_transitive(libd, [(liba, True, False, False, False),
                                 (libb, True, True, False, False),
                                 (libc, True, True, False, True)])
        _check_transitive(liba, [(libb, True, True, False, False),
                                 (libc, True, True, False, True)])

    def test_multiple_levels_transitive_headers(self):
        # app -> libcc0.1 -> libb0.1  -> liba0.1
        self.recipe_cache("liba/0.1")
        self.recipe_conanfile("libb/0.1", GenConanfile().with_package_type("static-library")
                                                        .with_requirement("liba/0.1",
                                                                          transitive_headers=True))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("static-library")
                                                        .with_requirement("libb/0.1",
                                                                          transitive_headers=True))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, True, True, False, False)])


class TestLinearFourLevels(GraphManagerTest):
    def test_default(self):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile())
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, True, True, False, False)])

    def test_negate_headers(self):
        # app -> libc/0.1 -> libb0.1  -(not headers)-> liba0.1
        # So nobody depends on the headers downstream
        self.recipe_conanfile("liba/0.1", GenConanfile())
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1", headers=False))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, True, False, False)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, False, True, False, False)])

    @parameterized.expand([("static-library", ),
                           ("shared-library", )])
    def test_libraries_transitive_headers(self, library_type):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # All with transitive_headers, the final application shoud get all headers
        # https://github.com/conan-io/conan/issues/12504
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type(library_type))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type(library_type)
                                            .with_requirement("liba/0.1", transitive_headers=True))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type(library_type)
                                            .with_requirement("libb/0.1", transitive_headers=True))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        if library_type == "shared-library":
            _check_transitive(app, [(libc, True, True, False, True),
                                    (libb, True, False, False, True),
                                    (liba, True, False, False, True)])
        else:  # Both static and unknown behave the same
            _check_transitive(app, [(libc, True, True, False, False),
                                    (libb, True, True, False, False),
                                    (liba, True, True, False, False)])

    def test_negate_libs(self):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # even if all are static, we want to disable the propagation of one static lib downstream
        # because only the headers are used
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("liba/0.1", libs=False))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, False, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, False, False, False)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, False, True, False, False),
                                (liba, False, False, False, False)])

    def test_disable_transitive_libs(self):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # even if all are static, we want to disable the propagation of one static lib downstream
        # Maybe we are re-archiving
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("liba/0.1", transitive_libs=False))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, False, False, False)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, False, True, False, False),
                                (liba, False, False, False, False)])

    def test_shared_depends_static_libraries(self):
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        # Emulate a re-packaging, re-archiving static library
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("shared-library")
                                            .with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libc, [(libb, True, True, False, True),
                                 (liba, False, False, False, False)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, False, True, False, True),
                                (liba, False, False, False, False)])

    def test_negate_run(self):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # even if all are shared, we want to disable the propagation of one shared lib downstream
        # because only the headers are used
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("shared-library"))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("shared-library")
                                            .with_requirement("liba/0.1", run=False))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("shared-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, True),
                                 (liba, False, False, False, False)])

        _check_transitive(app, [(libc, True, True, False, True),
                                (libb, False, False, False, True),
                                (liba, False, False, False, False)])

    def test_force_run(self):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # even if all are static, there is something in a static lib (files or whatever) that
        # is necessary at runtime
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("liba/0.1", run=True))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("static-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, True)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, True, False, True)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, False, True, False, False),
                                (liba, False, True, False, True)])

    @parameterized.expand([(True,),
                           (False,)])
    def test_header_only_run(self, run):
        # app -> libc/0.1 -> libb0.1  -> liba0.1
        # many header-onlys
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library"))
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_package_type("header-library")
                                            .with_requirement("liba/0.1", run=run))
        self.recipe_conanfile("libc/0.1",
                              GenConanfile().with_package_type("header-library")
                                            .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, False, False, run)])
        _check_transitive(libc, [(libb, True, False, False, False),
                                 (liba, True, False, False, run)])

        _check_transitive(app, [(libc, True, False, False, False),
                                (libb, True, False, False, False),
                                (liba, True, False, False, run)])

    def test_intermediate_header_only(self):
        # app -> libc/0.1 (static) -> libb0.1 (header) -> liba0.1 (static)
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",  GenConanfile().with_package_type("header-library")
                                                         .with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("static-library")
                                                        .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, False, False, False),
                                 (liba, True, True, False, False)])

        _check_transitive(app, [(libc, True, True, False, False),
                                (libb, False, False, False, False),
                                (liba, False, True, False, False)])

    def test_static_propagation_linear(self):
        # https://github.com/conan-io/conan/issues/16402
        # app -> libc/0.1 (shared) -> libb0.1 (static) -> liba0.1 (static)
        # The propagation of traits is correct, it is only the "skip-binaries" not able to skip
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",  GenConanfile().with_package_type("static-library")
                                                         .with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("shared-library")
                                                        .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, True, False, False)])
        _check_transitive(app, [(libc, True, True, False, True),
                                (libb, False, False, False, False),
                                (liba, False, False, False, False)])

    def test_incorrect_static_propagation_diamond(self):
        # https://github.com/conan-io/conan/issues/16402
        # app -> libc/0.1 (shared) -> libb0.1 (static) -> liba0.1 (static)
        #  \------------------------------------------------/
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1",  GenConanfile().with_package_type("static-library")
                                                         .with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_package_type("shared-library")
                                                        .with_requirement("libb/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1", "liba/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc, liba])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, app])

        # node, headers, lib, build, run
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, False, True, False, False)])
        _check_transitive(app, [(libc, True, True, False, True),
                                (libb, False, False, False, False),
                                (liba, True, True, False, False)])


class TestLinearFiveLevelsHeaders(GraphManagerTest):
    def test_all_header_only(self):
        # app -> libd/0.1 -> libc/0.1 -> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1")
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, False, False, False),
                                (libb, True, False, False, False),
                                (liba, True, False, False, False)])

    def test_all_header_only_aggregating_libc(self):
        # libc is copying and pasting the others headers at build time, creating re-distribution
        # app -> libd/0.1 -> libc/0.1 -(transitive_headers=FALSE)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("header-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          transitive_headers=False)
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, False, False, False),
                                (libb, False, False, False, False),
                                (liba, False, False, False, False)])

    def test_first_header_only(self):
        # app -> libd/0.1(header) -> libc/0.1 -> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, False, True, False, False),
                                (liba, False, True, False, False)])

    def test_first_header_only_transitive_headers_b(self):
        # app -> libd/0.1(header) -> libc/0.1 -(transitive_headers=T)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          transitive_headers=True)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, False, True, False, False)])

    def test_first_header_only_transitive_headers_b_a(self):
        # app -> libd/0.1(header) -> libc/0.1 -(transitive_headers=T)-> libb0.1 -(transitive_headers=T)-> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1",
                                                                          transitive_headers=True)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          transitive_headers=True)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, True, True, False, False)])

    def test_first_header_only_reject_libs(self):
        # Like Libc knows it only uses headers from libb
        # app -> libd/0.1(header) -> libc/0.1 -(libs=False)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1", libs=False)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, False, False, False, False),
                                (liba, False, False, False, False)])

    def test_d_b_header_only(self):
        # app -> libd/0.1(header) -> libc/0.1 -> libb0.1(header)  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, False, False, False, False),
                                (liba, False, True, False, False)])

    def test_d_b_header_only_transitive_headers_b(self):
        # app -> libd/0.1(header) -> libc/0.1 -(transitive_headers=T)-> libb0.1(header)  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("header-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          transitive_headers=True)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("header-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, False, False, False),
                                (libc, True, True, False, False),
                                (libb, True, False, False, False),
                                (liba, True, True, False, False)])

    def test_visible_transitivity(self):
        # app -> libd/0.1 -> libc/0.1 -(visible=False)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile())
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          visible=False))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False),
                                (libc, True, True, False, False)])
        _check_transitive(libd, [(libc, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, True, True, False, False)])

    def test_visible_build_transitivity(self):
        # app -> libd/0.1 -> libc/0.1 -(visible=True, build=True)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile())
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1", build=True))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False),
                                (libc, True, True, False, False),
                                (libb, False, False, True, False)])
        _check_transitive(libd, [(libc, True, True, False, False),
                                 (libb, False, False, True, False)])
        _check_transitive(libc, [(libb, True, True, True, False)])


class TestLinearFiveLevelsLibraries(GraphManagerTest):
    def test_all_static(self):
        # app -> libd/0.1 -> libc/0.1 -> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("static-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False),
                                (libc, False, True, False, False),
                                (libb, False, True, False, False),
                                (liba, False, True, False, False)])

    def test_libc_aggregating_static(self):
        # Lets think libc is re-linking its dependencies in a single .lib
        # app -> libd/0.1 -> libc/0.1 -(transitive_libs=False)-> libb0.1  -> liba0.1
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("static-library"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1")
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("libb/0.1",
                                                                          transitive_libs=False)
                                                        .with_package_type("static-library"))
        self.recipe_conanfile("libd/0.1", GenConanfile().with_requirement("libc/0.1")
                                                        .with_package_type("static-library"))
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = libd.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False),
                                (libc, False, True, False, False),
                                (libb, False, False, False, False),
                                (liba, False, False, False, False)])


class TestDiamond(GraphManagerTest):

    def test_diamond(self):
        # app -> libb0.1 -> liba0.1
        #    \-> libc0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        _check_transitive(app, [(libb, True, True, False, False),
                                (libc, True, True, False, False),
                                (liba, True, True, False, False)])

    @parameterized.expand([(True, ), (False, )])
    def test_diamond_additive(self, order):
        # app -> libb0.1 ---------> liba0.1
        #    \-> libc0.1 (run=True)->/
        self.recipe_cache("liba/0.1")
        if order:
            self.recipe_cache("libb/0.1", ["liba/0.1"])
            self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("liba/0.1", run=True))
        else:
            self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1", run=True))
            self.recipe_cache("libc/0.1", ["liba/0.1"])

        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        _check_transitive(app, [(libb, True, True, False, False),
                                (libc, True, True, False, False),
                                (liba, True, True, False, True)])

    def test_half_diamond(self):
        # app -----------> liba0.1
        #    \-> libc0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["liba/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        liba = app.edges[0].dst
        libc = app.edges[1].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[liba, libc])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[app, libc])

        # Order seems to be link order
        _check_transitive(app, [(libc, True, True, False, False),
                                (liba, True, True, False, False)])
        # Both requires of app are direct! https://github.com/conan-io/conan/pull/12388
        for require in app.transitive_deps.keys():
            assert require.direct is True

    def test_half_diamond_reverse(self):
        # same as above, just swap order of declaration
        # app -->libc0.1--> liba0.1
        #    \-------------->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1", "liba/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        liba = app.edges[1].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[liba, libc])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[app, libc])

        # Order seems to be link order, constant irrespective of declaration order, good
        _check_transitive(app, [(libc, True, True, False, False),
                                (liba, True, True, False, False)])
        # Both requires of app are direct! https://github.com/conan-io/conan/pull/12388
        for require in app.transitive_deps.keys():
            assert require.direct is True

    def test_shared_static(self):
        # app -> libb0.1 (shared) -> liba0.1 (static)
        #    \-> libc0.1 (shared) ->/
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=True)
        self.recipe_cache("libc/0.1", ["liba/0.1"], option_shared=True)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba = libb.edges[0].dst
        liba1 = libc.edges[0].dst

        assert liba is liba1

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, True),
                                (libc, True, True, False, True),
                                (liba, False, False, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(liba, True, True, False, False)])

    def test_private(self):
        # app -> libd0.1 -(private)-> libb0.1 -> liba0.1
        #            \ ---(private)-> libc0.1 --->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        libd = GenConanfile().with_requirement("libb/0.1", visible=False)
        libd.with_requirement("libc/0.1", visible=False)
        self.recipe_conanfile("libd/0.1", libd)
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libb = libd.edges[0].dst
        libc = libd.edges[1].dst
        liba = libb.edges[0].dst
        liba2 = libc.edges[0].dst

        assert liba is liba2

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libb, libc], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        # node, headers, lib, build, run
        _check_transitive(app, [(libd, True, True, False, False)])
        _check_transitive(libd, [(libb, True, True, False, False),
                                 (libc, True, True, False, False),
                                 (liba, True, True, False, False)])

    def test_shared_static_private(self):
        # app -> libb0.1 (shared) -(private)-> liba0.1 (static)
        #    \-> libc0.1 (shared) -> liba0.2 (static)
        # This private allows to avoid the liba conflict
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("liba/0.2", option_shared=False)
        libb = GenConanfile().with_requirement("liba/0.1", visible=False)
        libb.with_shared_option(True)
        self.recipe_conanfile("libb/0.1", libb)
        self.recipe_cache("libc/0.1", ["liba/0.2"], option_shared=True)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba1 = libb.edges[0].dst
        liba2 = libc.edges[0].dst

        assert liba1 is not liba2

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba2], dependents=[app])
        self._check_node(liba1, "liba/0.1#123", dependents=[libb])
        self._check_node(liba2, "liba/0.2#123", dependents=[libc])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, True),
                                (libc, True, True, False, True),
                                (liba2, False, False, False, False)])
        _check_transitive(libb, [(liba1, True, True, False, False)])
        _check_transitive(libc, [(liba2, True, True, False, False)])

    def test_diamond_conflict(self):
        # app -> libb0.1 -> liba0.1
        #    \-> libc0.1 -> liba0.2
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.2"])

        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphConflictError

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba1 = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[], dependents=[app])
        self._check_node(liba1, "liba/0.1#123", dependents=[libb])

    def test_shared_conflict_shared(self):
        # app -> libb0.1 (shared) -> liba0.1 (shared)
        #    \-> libc0.1 (shared) -> liba0.2 (shared)
        self.recipe_cache("liba/0.1", option_shared=True)
        self.recipe_cache("liba/0.2", option_shared=True)
        self.recipe_cache("libb/0.1", ["liba/0.1"], option_shared=True)
        self.recipe_cache("libc/0.1", ["liba/0.2"], option_shared=True)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphConflictError

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba1 = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[], dependents=[app])
        self._check_node(liba1, "liba/0.1#123", dependents=[libb])

    def test_private_conflict(self):
        # app -> libd0.1 -(private)-> libb0.1 -> liba0.1
        #            \ ---(private)-> libc0.1 -> liba0.2
        #
        # private requires do not avoid conflicts at the node level, only downstream
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.2"])
        libd = GenConanfile().with_requirement("libb/0.1", visible=False)
        libd.with_requirement("libc/0.1", visible=False)
        self.recipe_conanfile("libd/0.1", libd)
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1"])

        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphConflictError

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libb = libd.edges[0].dst
        libc = libd.edges[1].dst
        liba1 = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libd])
        self._check_node(libd, "libd/0.1#123", deps=[libb, libc], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[libd])
        self._check_node(libc, "libc/0.1#123", deps=[], dependents=[libd])
        self._check_node(liba1, "liba/0.1#123", dependents=[libb])

    def test_diamond_transitive_private(self):
        # https://github.com/conan-io/conan/issues/13630
        # libc0.1 ----------> liba0.1 --(private) -> zlib/1.0
        #      \-> --(libb---->/

        self.recipe_cache("zlib/0.1")
        self.recipe_conanfile("liba/0.1", GenConanfile("liba", "0.1")
                              .with_requirement("zlib/0.1", visible=False))
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1", "libb/0.1"])
        consumer = self.consumer_conanfile(GenConanfile("libc", "0.1")
                                           .with_requires("liba/0.1", "libb/0.1"))
        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        libc = deps_graph.root
        liba = libc.edges[0].dst
        libb = libc.edges[1].dst
        liba1 = libb.edges[0].dst
        zlib = liba.edges[0].dst

        assert liba is liba1
        # TODO: No Revision??? Because of consumer?
        self._check_node(libc, "libc/0.1", deps=[liba, libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc], deps=[zlib])
        self._check_node(zlib, "zlib/0.1#123", dependents=[liba])

        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, True, True, False, False)])

    def test_private_transitive_headers_no_conflict(self):
        # https://github.com/conan-io/conan/issues/15559
        # app -->liba/0.1 -(private)-> spdlog/0.1(header-only) -> fmt/0.1 (header-only)
        #   \ --------------------------------------------------> fmt/0.2
        self.recipe_conanfile("fmt/0.1",
                              GenConanfile("fmt", "0.1").with_package_type("header-library"))
        self.recipe_conanfile("fmt/0.2", GenConanfile("fmt", "0.2"))
        self.recipe_conanfile("spdlog/0.1",
                              GenConanfile("spdlog", "0.1").with_package_type("header-library")
                                                           .with_requires("fmt/0.1"))
        self.recipe_conanfile("liba/0.1",
                              GenConanfile("liba", "0.2").with_requirement("spdlog/0.1",
                                                                           visible=False))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1")
                                           .with_requires("liba/0.1", "fmt/0.2"))
        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        liba = app.edges[0].dst
        spdlog = liba.edges[0].dst
        fmt01 = spdlog.edges[0].dst
        fmt02 = app.edges[1].dst

        self._check_node(app, "app/0.1", deps=[liba, fmt02])
        self._check_node(liba, "liba/0.1#123", deps=[spdlog], dependents=[app])
        self._check_node(spdlog, "spdlog/0.1#123", deps=[fmt01], dependents=[liba])
        self._check_node(fmt01, "fmt/0.1#123", deps=[], dependents=[spdlog])
        self._check_node(fmt02, "fmt/0.2#123", dependents=[app])

        # node, headers, lib, build, run
        _check_transitive(app, [(liba, True, True, False, False),
                                (fmt02, True, True, False, False)])


class TestDiamondMultiple(GraphManagerTest):

    def test_consecutive_diamonds(self):
        # app -> libe0.1 -> libd0.1 -> libb0.1 -> liba0.1
        #    \-> libf0.1 ->/    \-> libc0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        self.recipe_cache("libd/0.1", ["libb/0.1", "libc/0.1"])
        self.recipe_cache("libe/0.1", ["libd/0.1"])
        self.recipe_cache("libf/0.1", ["libd/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libe/0.1", "libf/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 7 == len(deps_graph.nodes)
        app = deps_graph.root
        libe = app.edges[0].dst
        libf = app.edges[1].dst
        libd = libe.edges[0].dst
        libb = libd.edges[0].dst
        libc = libd.edges[1].dst
        liba = libc.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libe, libf])
        self._check_node(libe, "libe/0.1#123", deps=[libd], dependents=[app])
        self._check_node(libf, "libf/0.1#123", deps=[libd], dependents=[app])
        self._check_node(libd, "libd/0.1#123", deps=[libb, libc], dependents=[libe, libf])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        _check_transitive(app, [(libe, True, True, False, False),
                                (libf, True, True, False, False),
                                (libd, True, True, False, False),
                                (libb, True, True, False, False),
                                (libc, True, True, False, False),
                                (liba, True, True, False, False)])

    def test_consecutive_diamonds_private(self):
        # app -> libe0.1 ---------> libd0.1 ---> libb0.1 ---> liba0.1
        #    \-> (private)->libf0.1 ->/    \-private-> libc0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        self._cache_recipe("libd/0.1", GenConanfile().with_require("libb/0.1")
                           .with_requirement("libc/0.1", visible=False))
        self.recipe_cache("libe/0.1", ["libd/0.1"])
        self.recipe_cache("libf/0.1", ["libd/0.1"])
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libe/0.1")
                                           .with_requirement("libf/0.1", visible=False))

        deps_graph = self.build_consumer(consumer)

        assert 7 == len(deps_graph.nodes)
        app = deps_graph.root
        libe = app.edges[0].dst
        libf = app.edges[1].dst
        libd = libe.edges[0].dst
        libb = libd.edges[0].dst
        libc = libd.edges[1].dst
        liba = libc.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libe, libf])
        self._check_node(libe, "libe/0.1#123", deps=[libd], dependents=[app])
        self._check_node(libf, "libf/0.1#123", deps=[libd], dependents=[app])
        self._check_node(libd, "libd/0.1#123", deps=[libb, libc], dependents=[libe, libf])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libd])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        # FIXME: In this case the order seems a bit broken
        _check_transitive(app, [(libe, True, True, False, False),
                                (libf, True, True, False, False),
                                (libd, True, True, False, False),
                                (libb, True, True, False, False),
                                (liba, True, True, False, False),
                                ])

    def test_parallel_diamond(self):
        # app -> libb0.1 -> liba0.1
        #    \-> libc0.1 ->/
        #    \-> libd0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        self.recipe_cache("libd/0.1", ["liba/0.1"])

        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1", "libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        libd = app.edges[2].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb, libc, libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libd, "libd/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc, libd])

    def test_nested_diamond(self):
        # app --------> libb0.1 -> liba0.1
        #    \--------> libc0.1 ->/
        #     \-> libd0.1 ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1"])
        self.recipe_cache("libd/0.1", ["libc/0.1"])

        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1", "libd/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        libd = app.edges[2].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb, libc, libd])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba], dependents=[app, libd])
        self._check_node(libd, "libd/0.1#123", deps=[libc], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

    def test_multiple_transitive(self):
        # https://github.com/conanio/conan/issues/4720
        # app -> libb0.1  -> libc0.1 -> libd0.1
        #    \--------------->/          /
        #     \------------------------>/
        self.recipe_cache("libd/0.1")
        self.recipe_cache("libc/0.1", ["libd/0.1"])
        self.recipe_cache("libb/0.1", ["libc/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libd/0.1", "libc/0.1", "libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libd = app.edges[0].dst
        libc = app.edges[1].dst
        libb = app.edges[2].dst

        self._check_node(app, "app/0.1", deps=[libd, libc, libb])
        self._check_node(libd, "libd/0.1#123", dependents=[app, libc])
        self._check_node(libb, "libb/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libd], dependents=[app, libb])

    def test_loop(self):
        # app -> libc0.1 -> libb0.1 -> liba0.1 ->|
        #             \<-------------------------|
        self.recipe_cache("liba/0.1", ["libc/0.1"])
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["libb/0.1"])

        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer, install=False)
        # TODO: Better error modeling
        assert type(deps_graph.error) == GraphLoopError

        assert 4 == len(deps_graph.nodes)

        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", deps=[], dependents=[libb])


class TestTransitiveOverridesGraph(GraphManagerTest):

    def test_diamond(self):
        # app -> libb0.1 -> liba0.2 (overriden to lib0.2)
        #    \-> --------- ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libb/0.1")
                                           .with_requirement("liba/0.2", force=True))
        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = app.edges[1].dst
        liba2 = libb.edges[0].dst

        assert liba is liba2

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb, liba])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.2#123", dependents=[libb, app])

    def test_diamond_conflict(self):
        # app -> libb0.1 -> liba0.2 (overriden to lib0.2)
        #    \-> --------- ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "liba/0.2"])

        deps_graph = self.build_consumer(consumer, install=False)
        assert deps_graph.error is not False
        assert type(deps_graph.error) == GraphConflictError

        assert 2 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[], dependents=[app])

    def test_build_script_no_conflict(self):
        # app -> libb0.1 -> liba0.1 (build-scripts)
        #    \-> libc0.1 -> liba0.2
        self.recipe_conanfile("liba/0.1", GenConanfile().with_package_type("build-scripts"))
        self.recipe_conanfile("liba/0.2", GenConanfile())
        self.recipe_conanfile("libb/0.1",
                              GenConanfile().with_tool_requirement("liba/0.1", run=False))
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requirement("liba/0.2"))
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba1 = libb.edges[0].dst
        liba2 = libc.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[liba2], dependents=[app])
        self._check_node(liba1, "liba/0.1#123", dependents=[libb])
        self._check_node(liba2, "liba/0.2#123", dependents=[libc])

        # node, headers, lib, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (libc, True, True, False, False),
                                (liba2, True, True, False, False)])
        _check_transitive(libb, [(liba1, False, False, True, False)])
        _check_transitive(libc, [(liba2, True, True, False, False)])

    def test_diamond_reverse_order(self):
        # foo ---------------------------------> dep1/2.0
        #   \ -> dep2/1.0--(dep1/1.0 overriden)-->/
        self.recipe_cache("dep1/1.0")
        self.recipe_cache("dep1/2.0")
        self.recipe_cache("dep2/1.0", ["dep1/1.0"])
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1")
                                           .with_requirement("dep1/2.0", force=True)
                                           .with_requirement("dep2/1.0"))
        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        dep1 = app.edges[0].dst
        dep2 = app.edges[1].dst

        self._check_node(app, "app/0.1", deps=[dep1, dep2])
        self._check_node(dep1, "dep1/2.0#123", deps=[], dependents=[app, dep2])
        self._check_node(dep2, "dep2/1.0#123", deps=[dep1], dependents=[app])

    def test_diamond_reverse_order_conflict(self):
        # foo ---------------------------------> dep1/2.0
        #   \ -> dep2/1.0--(dep1/1.0 overriden)-->/
        self.recipe_cache("dep1/1.0")
        self.recipe_cache("dep1/2.0")
        self.recipe_cache("dep2/1.0", ["dep1/1.0"])
        consumer = self.recipe_consumer("app/0.1", ["dep1/2.0", "dep2/1.0"])
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphConflictError

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        dep1 = app.edges[0].dst
        dep2 = app.edges[1].dst
        self._check_node(app, "app/0.1", deps=[dep1, dep2])
        self._check_node(dep1, "dep1/2.0#123", deps=[], dependents=[app])
        # dep2 no dependency, it was not resolved due to conflict
        self._check_node(dep2, "dep2/1.0#123", deps=[], dependents=[app])

    def test_invisible_not_forced(self):
        # app -> libb0.1 -(visible=False)----> liba0.1 (NOT forced to lib0.2)
        #    \-> -----(force not used)-------> liba0.2
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1",
                                                                          visible=False))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libb/0.1")
                                           .with_requirement("liba/0.2", force=True))
        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba2 = app.edges[1].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb, liba2])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])
        self._check_node(liba2, "liba/0.2#123", dependents=[app])


class TestPureOverride(GraphManagerTest):

    def test_diamond(self):
        # app -> libb0.1 -> liba0.2 (overriden to lib0.2)
        #    \-> ---(override)------ ->/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libb/0.1")
                                           .with_requirement("liba/0.2", override=True))
        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.2#123", dependents=[libb])

    def test_discarded_override(self):
        # app ->---(override)------> liba0.2
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1")
                                           .with_requirement("liba/0.2", override=True))
        deps_graph = self.build_consumer(consumer)

        assert 1 == len(deps_graph.nodes)
        app = deps_graph.root
        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[])

    def test_invisible_not_overriden(self):
        # app -> libb0.1 -(visible=False)----> liba0.1 (NOT overriden to lib0.2)
        #    \-> -----(override not used)------->/
        self.recipe_cache("liba/0.1")
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requirement("liba/0.1",
                                                                          visible=False))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libb/0.1")
                                           .with_requirement("liba/0.2", override=True))
        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])

    def test_nested_overrides(self):
        # app -> libc0.1 ------> libb0.1 ------> liba0.1
        #  \          \-> --(override liba0.2)---->/
        #   \-> ---(override liba0.3)------------>/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("liba/0.3")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_conanfile("libc/0.1", GenConanfile("libc", "0.1")
                              .with_requirement("libb/0.1")
                              .with_requirement("liba/0.2", override=True))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libc/0.1")
                                           .with_requirement("liba/0.3", override=True))
        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.3#123", dependents=[libb])

    def test_override_solve_upstream_conflict(self):
        # app -> libc0.1 ------> libb0.1 ------> liba0.1
        #  \          \-> --(liba0.2 conflict)---->/
        #   \-> ---(override liba0.3)------------>/
        self.recipe_cache("liba/0.1")
        self.recipe_cache("liba/0.2")
        self.recipe_cache("liba/0.3")
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_conanfile("libc/0.1", GenConanfile("libc", "0.1")
                              .with_requires("libb/0.1", "liba/0.2"))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_require("libc/0.1")
                                           .with_requirement("liba/0.3", override=True))
        deps_graph = self.build_consumer(consumer)

        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        # TODO: No Revision??? Because of consumer?
        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb, liba], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.3#123", dependents=[libb, libc])

    def test_override_not_used(self):
        # https://github.com/conan-io/conan/issues/13630
        # libc0.1 ----------> liba0.1 --(override) -> zlib/1.0
        #      \-> --(libb---->/

        self.recipe_conanfile("liba/0.1", GenConanfile("liba", "0.1")
                              .with_requirement("zlib/0.1", override=True))
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["liba/0.1", "libb/0.1"])
        consumer = self.consumer_conanfile(GenConanfile("libc", "0.1")
                                           .with_requires("liba/0.1", "libb/0.1"))
        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        libc = deps_graph.root
        liba = libc.edges[0].dst
        libb = libc.edges[1].dst
        liba1 = libb.edges[0].dst

        assert liba is liba1
        # TODO: No Revision??? Because of consumer?
        self._check_node(libc, "libc/0.1", deps=[liba, libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc])

        _check_transitive(libb, [(liba, True, True, False, False)])
        _check_transitive(libc, [(libb, True, True, False, False),
                                 (liba, True, True, False, False)])


class TestPackageIDDeductions(GraphManagerTest):

    def test_static_dep_to_shared(self):
        # project -> app1 -> lib
        #    \---- > app2 --/

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_requirement("lib/0.1"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1")
                                      .with_requirement("app1/0.1", headers=False, libs=False,
                                                        build=False, run=True)
                                      .with_requirement("app2/0.1", headers=False, libs=False,
                                                        build=False, run=True)
                                      )

        assert 4 == len(deps_graph.nodes)
        project = deps_graph.root
        app1 = project.edges[0].dst
        app2 = project.edges[1].dst
        lib = app1.edges[0].dst
        lib2 = app2.edges[0].dst

        assert lib is lib2

        self._check_node(project, "project/0.1@", deps=[app1, app2], dependents=[])
        self._check_node(app1, "app1/0.1#123", deps=[lib], dependents=[project])
        self._check_node(app2, "app2/0.1#123", deps=[lib], dependents=[project])
        self._check_node(lib, "lib/0.1#123", deps=[], dependents=[app1, app2])

        # node, headers, lib, build, run
        _check_transitive(project, [(app1, False, False, False, True),
                                    (app2, False, False, False, True),
                                    (lib, False, False, False, False)])


class TestProjectApp(GraphManagerTest):
    """
    Emulating a project that can gather multiple applications and other resources and build a
    consistent graph, in which dependencies are same versions
    """
    def test_project_require_transitive(self):
        # project -> app1 -> lib
        #    \---- > app2 --/

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_requirement("lib/0.1"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1")
                                      .with_requirement("app1/0.1", headers=False, libs=False,
                                                        build=False, run=True)
                                      .with_requirement("app2/0.1", headers=False, libs=False,
                                                        build=False, run=True)
                                      )

        assert 4 == len(deps_graph.nodes)
        project = deps_graph.root
        app1 = project.edges[0].dst
        app2 = project.edges[1].dst
        lib = app1.edges[0].dst
        lib2 = app2.edges[0].dst

        assert lib is lib2

        self._check_node(project, "project/0.1@", deps=[app1, app2], dependents=[])
        self._check_node(app1, "app1/0.1#123", deps=[lib], dependents=[project])
        self._check_node(app2, "app2/0.1#123", deps=[lib], dependents=[project])
        self._check_node(lib, "lib/0.1#123", deps=[], dependents=[app1, app2])

        # node, headers, lib, build, run
        _check_transitive(project, [(app1, False, False, False, True),
                                    (app2, False, False, False, True),
                                    (lib, False, False, False, False)])

    def test_project_require_transitive_conflict(self):
        # project -> app1 -> lib/0.1
        #    \---- > app2 -> lib/0.2

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("lib/0.2", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_requirement("lib/0.2"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1")
                                      .with_requirement("app1/0.1", headers=False, libs=False,
                                                        build=False, run=True)
                                      .with_requirement("app2/0.1", headers=False, libs=False,
                                                        build=False, run=True),
                                      install=False)

        assert type(deps_graph.error) == GraphConflictError

    def test_project_require_apps_transitive(self):
        # project -> app1 (app type) -> lib
        #    \---- > app2 (app type) --/

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_package_type("application").
                           with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_package_type("application").
                           with_requirement("lib/0.1"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1").with_requires("app1/0.1",
                                                                                   "app2/0.1"))

        assert 4 == len(deps_graph.nodes)
        project = deps_graph.root
        app1 = project.edges[0].dst
        app2 = project.edges[1].dst
        lib = app1.edges[0].dst
        lib2 = app2.edges[0].dst

        assert lib is lib2

        self._check_node(project, "project/0.1@", deps=[app1, app2], dependents=[])
        self._check_node(app1, "app1/0.1#123", deps=[lib], dependents=[project])
        self._check_node(app2, "app2/0.1#123", deps=[lib], dependents=[project])
        self._check_node(lib, "lib/0.1#123", deps=[], dependents=[app1, app2])

        # node, headers, lib, build, run
        _check_transitive(project, [(app1, False, False, False, True),
                                    (app2, False, False, False, True),
                                    (lib, False, False, False, False)])

    def test_project_require_apps_transitive_conflict(self):
        # project -> app1 (app type) -> lib/0.1
        #    \---- > app2 (app type) -> lib/0.2

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("lib/0.2", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_package_type("application").
                           with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_package_type("application").
                           with_requirement("lib/0.2"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1").with_requires("app1/0.1",
                                                                                   "app2/0.1"),
                                      install=False)

        assert type(deps_graph.error) == GraphConflictError

    def test_project_require_private(self):
        # project -(!visible)-> app1 -> lib1
        #    \----(!visible)- > app2 -> lib2
        # This doesn't conflict on project, as lib1, lib2 do not include, link or public

        self._cache_recipe("lib/0.1", GenConanfile())
        self._cache_recipe("lib/0.2", GenConanfile())
        self._cache_recipe("app1/0.1", GenConanfile().with_requirement("lib/0.1"))
        self._cache_recipe("app2/0.1", GenConanfile().with_requirement("lib/0.2"))

        deps_graph = self.build_graph(GenConanfile("project", "0.1")
                                      .with_requirement("app1/0.1", headers=False, libs=False,
                                                        build=False, run=True, visible=False)
                                      .with_requirement("app2/0.1", headers=False, libs=False,
                                                        build=False, run=True, visible=False)
                                      )

        assert 5 == len(deps_graph.nodes)
        project = deps_graph.root
        app1 = project.edges[0].dst
        app2 = project.edges[1].dst
        lib1 = app1.edges[0].dst
        lib2 = app2.edges[0].dst

        assert lib1 is not lib2

        self._check_node(project, "project/0.1@", deps=[app1, app2], dependents=[])
        self._check_node(app1, "app1/0.1#123", deps=[lib1], dependents=[project])
        self._check_node(app2, "app2/0.1#123", deps=[lib2], dependents=[project])
        self._check_node(lib1, "lib/0.1#123", deps=[], dependents=[app1])
        self._check_node(lib2, "lib/0.2#123", deps=[], dependents=[app2])

        # node, headers, lib, build, run
        _check_transitive(project, [(app1, False, False, False, True),
                                    (lib1, False, False, False, False),
                                    (app2, False, False, False, True),
                                    (lib2, False, False, False, False)])



import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.tool("cmake")
def test_build_modules_alias_target():
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy

        class Conan(ConanFile):
            name = "hello"
            version = "1.0"
            settings = "os", "arch", "compiler", "build_type"
            exports_sources = ["target-alias.cmake"]

            def package(self):
                copy(self, "target-alias.cmake", self.source_folder,
                     os.path.join(self.package_folder, "share/cmake"))

            def package_info(self):
                module = os.path.join("share", "cmake", "target-alias.cmake")
                self.cpp_info.set_property("cmake_build_modules", [module])
        """)

    target_alias = textwrap.dedent("""
        add_library(otherhello INTERFACE IMPORTED)
        target_link_libraries(otherhello INTERFACE hello::hello)
        """)
    client.save({"conanfile.py": conanfile, "target-alias.cmake": target_alias})
    client.run("create .")

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Conan(ConanFile):
            name = "consumer"
            version = "1.0"
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = ["CMakeLists.txt"]
            generators = "CMakeDeps", "CMakeToolchain"
            requires = "hello/1.0"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)
    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.0)
        project(test NONE)
        find_package(hello CONFIG)
        get_target_property(tmp otherhello INTERFACE_LINK_LIBRARIES)
        message("otherhello link libraries: ${tmp}")
        """)
    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    client.run("create .")
    assert "otherhello link libraries: hello::hello" in client.out


@pytest.mark.tool("cmake")
def test_build_modules_custom_script():
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy

        class Conan(ConanFile):
            name = "myfunctions"
            version = "1.0"
            exports_sources = ["*.cmake"]

            def package(self):
                copy(self, "*.cmake", self.source_folder, self.package_folder)

            def package_info(self):
                self.cpp_info.set_property("cmake_build_modules", ["myfunction.cmake"])
        """)

    myfunction = textwrap.dedent("""
        function(myfunction)
            message("Hello myfunction!!!!")
        endfunction()
        """)
    client.save({"conanfile.py": conanfile, "myfunction.cmake": myfunction})
    client.run("create .")

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeDeps

        class Conan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            tool_requires = "myfunctions/1.0"

            def generate(self):
                deps = CMakeDeps(self)
                deps.build_context_activated = ["myfunctions"]
                deps.build_context_build_modules = ["myfunctions"]
                deps.set_property("myfunctions", "cmake_find_mode", "module", build_context=True)
                deps.generate()

            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)
    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.0)
        project(test NONE)
        find_package(myfunctions MODULE REQUIRED)
        myfunction()
        """)
    client.save({"conanfile.py": consumer,
                 "CMakeLists.txt": cmakelists},
                clean_first=True)
    client.run("build .")
    assert "Hello myfunction!!!!" in client.out


@pytest.mark.tool("cmake")
def test_build_modules_components_is_not_possible():
    """
    The "cmake_build_module" property declared in the components is useless
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy

        class Conan(ConanFile):
            name = "openssl"
            version = "1.0"
            settings = "os", "arch", "compiler", "build_type"
            exports_sources = ["crypto.cmake", "root.cmake"]

            def package(self):
                copy(self, "*.cmake", self.source_folder, os.path.join(self.package_folder, "share/cmake"))

            def package_info(self):
                crypto_module = os.path.join("share", "cmake", "crypto.cmake")
                self.cpp_info.components["crypto"].set_property("cmake_build_modules", [crypto_module])

                root_module = os.path.join("share", "cmake", "root.cmake")
                self.cpp_info.set_property("cmake_build_modules", [root_module])
        """)

    crypto_cmake = textwrap.dedent("""
        function(crypto_message MESSAGE_OUTPUT)
            message("CRYPTO MESSAGE:${ARGV${0}}")
        endfunction()
        """)
    root_cmake = textwrap.dedent("""
        function(root_message MESSAGE_OUTPUT)
            message("ROOT MESSAGE:${ARGV${0}}")
        endfunction()
        """)
    client.save({"conanfile.py": conanfile,
                 "crypto.cmake": crypto_cmake,
                 "root.cmake": root_cmake})
    client.run("create .")

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Conan(ConanFile):
            name = "consumer"
            version = "1.0"
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = ["CMakeLists.txt"]
            generators = "CMakeDeps", "CMakeToolchain"
            requires = "openssl/1.0"

            def build(self):
                cmake = CMake(self)
                cmake.configure()

            def package_info(self):
                self.cpp_info.requires = ["openssl::crypto"]
        """)

    cmakelists = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.0)
            project(test NONE)
            find_package(openssl CONFIG)
            crypto_message("hello!")
            root_message("hello!")
            """)
    client.save({"conanfile.py": consumer, "CMakeLists.txt": cmakelists})
    # As we are requiring only "crypto" but it doesn't matter, it is not possible to include
    # only crypto build_modules
    client.run("create .", assert_error=True)
    assert 'Unknown CMake command "crypto_message"' in client.out

    # Comment the function call
    client.save({"CMakeLists.txt": cmakelists.replace("crypto", "#crypto")})
    assert "ROOT MESSAGE:hello!" not in client.out


@pytest.mark.tool("cmake")
@pytest.mark.parametrize("editable", [True, False])
def test_build_modules_custom_script_editable(editable):
    c = TestClient()
    conanfile = textwrap.dedent(r"""
        import os, glob
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        from conan.tools.files import copy, save

        class Conan(ConanFile):
            name = "myfunctions"
            version = "1.0"
            exports_sources = ["src/*.cmake"]
            settings = "build_type", "arch"

            def build(self):
                cmake = 'set(MY_CMAKE_PATH ${CMAKE_CURRENT_LIST_DIR})\n'\
                        'macro(otherfunc)\n'\
                        'file(READ "${MY_CMAKE_PATH}/my.txt" c)\n'\
                        'message("Hello ${c}!!!!")\nendmacro()'
                save(self, "otherfuncs.cmake", cmake)
                save(self, "my.txt", "contents of text file!!!!")

            def layout(self):
                cmake_layout(self, src_folder="src")
                src = glob.glob(os.path.join(self.recipe_folder, self.folders.source, "*.cmake"))
                build = glob.glob(os.path.join(self.recipe_folder, self.folders.build, "*.cmake"))
                self.cpp.source.set_property("cmake_build_modules", src)
                self.cpp.build.set_property("cmake_build_modules", build)

            def package(self):
                copy(self, "*.cmake", self.source_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)
                copy(self, "*.cmake", self.build_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)
                copy(self, "*.txt", self.build_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)

            def package_info(self):
                self.cpp_info.set_property("cmake_build_modules", glob.glob("mods/*.cmake"))
        """)

    myfunction = textwrap.dedent("""
        function(myfunction)
            message("Hello myfunction!!!!")
        endfunction()
        """)

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Conan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain", "CMakeDeps"
            requires = "myfunctions/1.0"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)
    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(test NONE)
        find_package(myfunctions CONFIG REQUIRED)
        myfunction()
        otherfunc()
        """)
    c.save({"functions/conanfile.py": conanfile,
            "functions/src/myfunction.cmake": myfunction,
            "app/conanfile.py": consumer,
            "app/CMakeLists.txt": cmakelists})

    if editable:
        c.run("editable add functions")
        c.run('build functions -c tools.cmake.cmake_layout:build_folder_vars="[\'settings.arch\']"')
    else:
        c.run("create functions")
    c.run('build app -c tools.cmake.cmake_layout:build_folder_vars="[\'settings.arch\']"')
    assert "Hello myfunction!!!!" in c.out
    assert "Hello contents of text file!!!!" in c.out


@pytest.mark.tool("cmake")
@pytest.mark.parametrize("editable", [True, False])
def test_build_modules_custom_script_editable_package(editable):
    # same as above, but with files from the package folder
    c = TestClient()
    conanfile = textwrap.dedent(r"""
        import os, glob
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        from conan.tools.files import copy, save

        class Conan(ConanFile):
            name = "myfunctions"
            version = "1.0"
            exports_sources = ["src/*.cmake", "src/vis/*.vis"]
            settings = "build_type", "arch"

            def build(self):
                config = str(self.settings.build_type).upper()
                cmake = f'set(MY_CMAKE_PATH ${{{self.name}_RES_DIRS_{config}}})\n'\
                        'macro(otherfunc)\n'\
                        'file(READ "${MY_CMAKE_PATH}/my.vis" c)\n'\
                        'message("Hello ${c}!!!!")\nendmacro()'
                save(self, "otherfuncs.cmake", cmake)

            def layout(self):
                cmake_layout(self, src_folder="src")
                src = glob.glob(os.path.join(self.recipe_folder, self.folders.source, "*.cmake"))
                build = glob.glob(os.path.join(self.recipe_folder, self.folders.build, "*.cmake"))
                self.cpp.source.set_property("cmake_build_modules", src)
                self.cpp.build.set_property("cmake_build_modules", build)
                self.cpp.source.resdirs = ["vis"]

            def package(self):
                copy(self, "*.cmake", self.source_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)
                copy(self, "*.cmake", self.build_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)
                copy(self, "*.vis", self.source_folder, os.path.join(self.package_folder, "mods"),
                     keep_path=False)

            def package_info(self):
                self.cpp_info.resdirs = ["mods"]
                self.cpp_info.set_property("cmake_build_modules", glob.glob("mods/*.cmake"))
        """)

    myfunction = textwrap.dedent("""
        function(myfunction)
            message("Hello myfunction!!!!")
        endfunction()
        """)

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Conan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain", "CMakeDeps"
            requires = "myfunctions/1.0"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
        """)
    cmakelists = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(test NONE)
        find_package(myfunctions CONFIG REQUIRED)
        myfunction()
        otherfunc()
        """)
    c.save({"functions/conanfile.py": conanfile,
            "functions/src/myfunction.cmake": myfunction,
            "functions/src/vis/my.vis": "contents of text file!!!!",
            "app/conanfile.py": consumer,
            "app/CMakeLists.txt": cmakelists})

    if editable:
        c.run("editable add functions")
        c.run('build functions -c tools.cmake.cmake_layout:build_folder_vars="[\'settings.arch\']"')
    else:
        c.run("create functions -vv")
    c.run('build app -c tools.cmake.cmake_layout:build_folder_vars="[\'settings.arch\']"')
    assert "Hello myfunction!!!!" in c.out
    assert "Hello contents of text file!!!!" in c.out



import re
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_require_different_versions():
    """ this test demostrates that it is possible to tool_require different versions
    of the same thing, deactivating run=False (as long as their executables are not called the same)

    https://github.com/conan-io/conan/issues/13521
    """
    c = TestClient()
    gcc = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            name = "gcc"
            def package(self):
                echo = f"@echo off\necho MYGCC={self.version}!!"
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.version}.bat"), echo)
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.version}.sh"), echo)
                os.chmod(os.path.join(self.package_folder, "bin", f"mygcc{self.version}.sh"), 0o777)
            """)
    wine = textwrap.dedent("""
        import os, platform
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            name = "wine"
            version = "1.0"
            def build_requirements(self):
                self.tool_requires("gcc/1.0", run=False)
                self.tool_requires("gcc/2.0", run=False)

            def generate(self):
                gcc1 = self.dependencies.build["gcc/1.0"]
                assert gcc1.ref.version == "1.0"
                gcc2 = self.dependencies.build["gcc/2.0"]
                assert gcc2.ref.version == "2.0"

            def build(self):
                ext = "bat" if platform.system() == "Windows" else "sh"
                self.run(f"mygcc1.0.{ext}")
                self.run(f"mygcc2.0.{ext}")
            """)
    c.save({"gcc/conanfile.py": gcc,
            "wine/conanfile.py": wine})

    c.run("create gcc --version=1.0")
    c.run("create gcc --version=2.0")

    c.run("build wine --lockfile-out=conan.lock")
    assert "gcc/1.0#3d6110b9e2b90074160fa33b6f0ea968 - Cache" in c.out
    assert "gcc/2.0#3d6110b9e2b90074160fa33b6f0ea968 - Cache" in c.out
    assert "MYGCC=1.0!!" in c.out
    assert "MYGCC=2.0!!" in c.out
    lock = c.load("wine/conan.lock")
    assert "gcc/1.0#3d6110b9e2b90074160fa33b6f0ea968" in lock
    assert "gcc/2.0#3d6110b9e2b90074160fa33b6f0ea968" in lock


def test_require_different_versions_profile_override():
    """ same as above but what if the profile is the one overriding the version?
    """
    c = TestClient()
    gcc = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            name = "gcc"
            def package(self):
                echo = f"@echo off\necho MYGCC={self.version}!!"
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.version}.bat"), echo)
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.version}.sh"), echo)
                os.chmod(os.path.join(self.package_folder, "bin", f"mygcc{self.version}.sh"), 0o777)
            """)
    wine = textwrap.dedent("""
        import os, platform
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            name = "wine"
            version = "1.0"
            def build_requirements(self):
                self.tool_requires("gcc/1.0", run=False)

            def build(self):
                ext = "bat" if platform.system() == "Windows" else "sh"
                self.run(f"mygcc1.0.{ext}")
                self.run(f"mygcc2.0.{ext}")
            """)
    c.save({"gcc/conanfile.py": gcc,
            "wine/conanfile.py": wine,
            "profile": "[tool_requires]\ngcc/2.0"})

    c.run("create gcc --version=1.0")
    c.run("create gcc --version=2.0")

    c.run("build wine -pr=profile --lockfile-out=conan.lock")
    assert "gcc/1.0#3d6110b9e2b90074160fa33b6f0ea968 - Cache" in c.out
    assert "gcc/2.0#3d6110b9e2b90074160fa33b6f0ea968 - Cache" in c.out
    assert "MYGCC=1.0!!" in c.out
    assert "MYGCC=2.0!!" in c.out
    lock = c.load("wine/conan.lock")
    assert "gcc/1.0#3d6110b9e2b90074160fa33b6f0ea968" in lock
    assert "gcc/2.0#3d6110b9e2b90074160fa33b6f0ea968" in lock


def test_require_different_versions_profile_override_build_script():
    """ build-scripts by default do the right thing, because they have run=True
    (they could be runnable shell scripts)
    """
    c = TestClient(light=True)
    buildscripts = GenConanfile("buildscripts").with_package_type("build-scripts")
    wine = GenConanfile("wine", "1.0").with_tool_requirement("buildscripts/1.0")
    c.save({"buildscripts/conanfile.py": buildscripts,
            "wine/conanfile.py": wine,
            "profile": "[tool_requires]\nbuildscripts/2.0"})

    c.run("create buildscripts --version=2.0")

    c.run("build wine -pr=profile --lockfile-out=conan.lock")
    assert "buildscripts/1.0" not in c.out
    assert "buildscripts/2.0#fced952ee7aba96f858b70c7d6c9c8d2 - Cache" in c.out
    lock = c.load("wine/conan.lock")
    assert "buildscripts/1.0" not in lock
    assert "buildscripts/2.0#fced952ee7aba96f858b70c7d6c9c8d2" in lock


def test_require_different_options():
    """ this test demostrates that it is possible to tool_require different options
    of the same thing, deactivating run=False (as long as their executables are not called the same)

    https://github.com/conan-io/conan/issues/13521
    """
    c = TestClient()
    gcc = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            name = "gcc"
            version = "1.0"
            options = {"myoption": [1, 2]}
            def package(self):
                echo = f"@echo off\necho MYGCC={self.options.myoption}!!"
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.options.myoption}.bat"), echo)
                save(self, os.path.join(self.package_folder, "bin", f"mygcc{self.options.myoption}.sh"), echo)
                os.chmod(os.path.join(self.package_folder, "bin", f"mygcc{self.options.myoption}.sh"), 0o777)
            """)
    wine = textwrap.dedent("""
        import os, platform
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            name = "wine"
            version = "1.0"
            def build_requirements(self):
                self.tool_requires("gcc/1.0", run=False, options={"myoption": 1})
                self.tool_requires("gcc/1.0", run=False, options={"myoption": 2})

            def generate(self):
                gcc1 = self.dependencies.build.get("gcc", options={"myoption": 1})
                assert gcc1.options.myoption == "1"
                gcc2 = self.dependencies.build.get("gcc", options={"myoption": 2})
                assert gcc2.options.myoption == "2"

            def build(self):
                ext = "bat" if platform.system() == "Windows" else "sh"
                self.run(f"mygcc1.{ext}")
                self.run(f"mygcc2.{ext}")
            """)
    c.save({"gcc/conanfile.py": gcc,
            "wine/conanfile.py": wine})

    c.run("create gcc -o myoption=1")
    c.run("create gcc -o myoption=2")

    c.run("build wine --lockfile-out=conan.lock")
    assert "gcc/1.0#616ce3babcecef39a27806c1a5f4b4ff - Cache" in c.out
    assert "MYGCC=1!!" in c.out
    assert "MYGCC=2!!" in c.out
    lock = c.load("wine/conan.lock")
    # Testing it doesn't crash or anything like that
    assert "gcc/1.0#616ce3babcecef39a27806c1a5f4b4ff" in lock


def test_require_different_versions_transitive():
    """
    https://github.com/conan-io/conan/issues/18086
    """
    c = TestClient(default_server_user=True, path_with_spaces=False)
    qemu = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            name = "myqemu"
            package_type = "application"

            def package(self):
                echo = f"@echo off\necho RUNNING {self.name}/{self.version}!!"
                save(self, os.path.join(self.package_folder, "bin", f"{self.name}.bat"), echo)
                save(self, os.path.join(self.package_folder, "bin", f"{self.name}.sh"), echo)
                os.chmod(os.path.join(self.package_folder, "bin", f"{self.name}.sh"), 0o777)
            """)
    mytool = textwrap.dedent(r"""
        import os, platform
        from conan import ConanFile
        from conan.tools.files import save, chdir
        class Pkg(ConanFile):
            version = "1.0"
            package_type = "application"

            def requirements(self):
                version = "1.0" if self.name == "snippy" else "2.0"
                self.requires(f"myqemu/{version}", visible=False, no_skip=True)

            def package(self):
                c = f'call "%1/myqemu.bat"' if platform.system() == "Windows" else f'"$1/myqemu.sh"'
                echo = f"@echo off\necho RUNNING {self.name}/{self.version}!!\n{c}"
                save(self, os.path.join(self.package_folder, "bin", f"{self.name}.bat"), echo)
                save(self, os.path.join(self.package_folder, "bin", f"{self.name}"), echo)
                os.chmod(os.path.join(self.package_folder, "bin", f"{self.name}"), 0o777)

            def package_info(self):
                pf = self.dependencies["myqemu"].cpp_info.bindir.replace("\\", "/")
                self.conf_info.define_path(f"user.myorg:{self.name}_qemu", pf)
            """)
    consumer = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            name = "consumer"
            version = "1.0"
            def build_requirements(self):
                self.tool_requires("snippy/1.0")
                self.tool_requires("valgrind/1.0")
            def build(self):
                qemu_snippy = self.conf.get("user.myorg:snippy_qemu")
                qemu_valgrind = self.conf.get("user.myorg:valgrind_qemu")
                self.run(f"valgrind {qemu_valgrind}")
                self.run(f'snippy {qemu_snippy}')
        """)

    c.save({"qemu/conanfile.py": qemu,
            "tool/conanfile.py": mytool,
            "consumer/conanfile.py": consumer})

    c.run("create qemu --version=1.0")
    c.run("create qemu --version=2.0")
    c.run("create tool --name=snippy")
    c.run("create tool --name=valgrind")
    c.run("build consumer")
    assert "RUNNING valgrind/1.0!!" in c.out
    assert "RUNNING myqemu/2.0!!" in c.out
    assert "RUNNING snippy/1.0!!" in c.out
    assert "RUNNING myqemu/1.0!!" in c.out

    c.run("upload * -r=default -c")
    # The "tools.graph:skip_binaries" shouldn't affect the result, it is never skipped
    for skip in ("-c tools.graph:skip_binaries=True", "-c tools.graph:skip_binaries=False", ""):
        c.run("remove * -c")
        # Re-downloads and it works
        c.run(f"build consumer {skip}")
        assert "RUNNING valgrind/1.0!!" in c.out
        assert "RUNNING myqemu/2.0!!" in c.out
        assert "RUNNING snippy/1.0!!" in c.out
        assert "RUNNING myqemu/1.0!!" in c.out

    c.run("create consumer")
    c.run("upload consumer/1.0 -r=default -c")
    c.run("remove * -c")
    c.run("install --requires=consumer/1.0")
    assert re.search(r"Skipped binaries(\s*)myqemu/1.0, myqemu/2.0, snippy/1.0, valgrind/1.0", c.out)



from unittest import mock

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.test.utils.env import environment_update


def test_package_revision_latest():
    # https://github.com/conan-io/conan/issues/14945
    c = TestClient()
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1").with_package_file("file.txt",
                                                                               env_var="MYVAR"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_tool_requires("tool/0.1")})
    with environment_update({"MYVAR": "MYVALUE1"}):
        # Just in case dates were not correctly processed
        with mock.patch("conan.internal.cache.cache.revision_timestamp_now",
                        return_value="1691760295"):
            c.run("create tool")
    prev1 = c.created_package_revision("tool/0.1")
    with environment_update({"MYVAR": "MYVALUE2"}):
        # Just in case dates were not correctly processed
        with mock.patch("conan.internal.cache.cache.revision_timestamp_now",
                        return_value="1697442658"):
            c.run("create tool")
    prev2 = c.created_package_revision("tool/0.1")

    c.run("create pkg")
    # The latest is used
    assert prev2 in c.out
    assert prev1 not in c.out

    c.run("install --requires=pkg/0.1 --build=pkg*")
    # The latest is used
    assert prev2 in c.out
    assert prev1 not in c.out



import os
from os.path import join, normpath, relpath

from conan.internal.errors import RecipeNotFoundException, PackageNotFoundException
from conan.internal.paths import CONAN_MANIFEST
from conan.api.model import PkgReference
from conan.api.model import RecipeReference
from conans.server.revision_list import RevisionList

REVISIONS_FILE = "revisions.txt"
SERVER_EXPORT_FOLDER = "export"
SERVER_PACKAGES_FOLDER = "package"


def ref_dir_repr(ref):
    return "/".join([ref.name, str(ref.version), ref.user or "_", ref.channel or "_"])


class ServerStore(object):

    def __init__(self, storage_adapter):
        self._storage_adapter = storage_adapter
        self._store_folder = storage_adapter._store_folder

    @property
    def store(self):
        return self._store_folder

    def base_folder(self, ref):
        assert ref.revision is not None, "BUG: server store needs RREV to get recipe reference"
        tmp = normpath(join(self.store, ref_dir_repr(ref)))
        return join(tmp, ref.revision)

    def conan_revisions_root(self, ref):
        """Parent folder of the conan package, for all the revisions"""
        assert not ref.revision, "BUG: server store doesn't need RREV to conan_revisions_root"
        return normpath(join(self.store, ref_dir_repr(ref)))

    def packages(self, ref):
        return join(self.base_folder(ref), SERVER_PACKAGES_FOLDER)

    def package(self, pref):
        assert pref.revision is not None, "BUG: server store needs PREV for package"
        tmp = join(self.packages(pref.ref), pref.package_id)
        return join(tmp, pref.revision)

    def export(self, ref):
        return join(self.base_folder(ref), SERVER_EXPORT_FOLDER)

    def get_recipe_file_path(self, ref, filename):
        abspath = join(self.export(ref), filename)
        return abspath

    def get_package_file_path(self, pref, filename):
        p_path = self.package(pref)
        abspath = join(p_path, filename)
        return abspath

    def path_exists(self, path):
        return self._storage_adapter.path_exists(path)

    # ############ ONLY FILE LIST SNAPSHOTS (APIv2)
    def get_recipe_file_list(self, ref):
        """Returns a  [filepath] """
        assert isinstance(ref, RecipeReference)
        files = self._get_file_list(self.export(ref))
        if CONAN_MANIFEST not in files:
            raise RecipeNotFoundException(ref)
        return files

    def get_package_file_list(self, pref):
        """Returns a  [filepath] """
        assert isinstance(pref, PkgReference)
        files = self._get_file_list(self.package(pref))
        if CONAN_MANIFEST not in files:
            raise PackageNotFoundException(pref)
        return files

    def _get_file_list(self, relative_path):
        file_list = self._storage_adapter.get_file_list(relative_path)
        file_list = [relpath(old_key, relative_path) for old_key in file_list]
        return file_list

    def _delete_empty_dirs(self, ref):
        lock_files = {REVISIONS_FILE, "%s.lock" % REVISIONS_FILE}

        ref_path = normpath(join(self.store, ref_dir_repr(ref)))
        if ref.revision:
            ref_path = join(ref_path, ref.revision)
        for _ in range(4 if not ref.revision else 5):
            if os.path.exists(ref_path):
                if set(os.listdir(ref_path)) == lock_files:
                    for lock_file in lock_files:
                        os.unlink(os.path.join(ref_path, lock_file))
                try:  # Take advantage that os.rmdir does not delete non-empty dirs
                    os.rmdir(ref_path)
                except OSError:
                    break  # not empty
            ref_path = os.path.dirname(ref_path)

    # ######### DELETE (APIv1 and APIv2)
    def remove_recipe(self, ref):
        assert isinstance(ref, RecipeReference)
        if not ref.revision:
            self._storage_adapter.delete_folder(self.conan_revisions_root(ref))
        else:
            self._storage_adapter.delete_folder(self.base_folder(ref))
            self._remove_revision_from_index(ref)
        self._delete_empty_dirs(ref)

    def remove_package(self, pref):
        assert isinstance(pref, PkgReference)
        assert pref.revision is not None, "BUG: server store needs PREV remove_package"
        assert pref.ref.revision is not None, "BUG: server store needs RREV remove_package"
        package_folder = self.package(pref)
        self._storage_adapter.delete_folder(package_folder)
        self._remove_package_revision_from_index(pref)

    def remove_all_packages(self, ref):
        assert ref.revision is not None, "BUG: server store needs RREV remove_all_packages"
        assert isinstance(ref, RecipeReference)
        packages_folder = self.packages(ref)
        self._storage_adapter.delete_folder(packages_folder)

    # Methods to manage revisions
    def get_last_revision(self, ref):
        assert(isinstance(ref, RecipeReference))
        rev_file_path = self._recipe_revisions_file(ref)
        return self._get_latest_revision(rev_file_path)

    def get_recipe_revisions_references(self, ref):
        """Returns a RevisionList"""
        if ref.revision:
            tmp = RevisionList()
            tmp.add_revision(ref.revision)
            return tmp.as_list()
        rev_file_path = self._recipe_revisions_file(ref)
        revs = self._get_revisions_list(rev_file_path).as_list()
        if not revs:
            raise RecipeNotFoundException(ref)
        return revs

    def get_last_package_revision(self, pref):
        assert(isinstance(pref, PkgReference))
        rev_file_path = self._package_revisions_file(pref)
        rev = self._get_latest_revision(rev_file_path)
        if rev:
            return PkgReference(pref.ref, pref.package_id, rev.revision, rev.time)
        return None

    def update_last_revision(self, ref):
        assert(isinstance(ref, RecipeReference))
        rev_file_path = self._recipe_revisions_file(ref)
        self._update_last_revision(rev_file_path, ref)

    def update_last_package_revision(self, pref):
        assert(isinstance(pref, PkgReference))
        rev_file_path = self._package_revisions_file(pref)
        self._update_last_revision(rev_file_path, pref)

    def _update_last_revision(self, rev_file_path, ref):
        if self._storage_adapter.path_exists(rev_file_path):
            rev_file = self._storage_adapter.read_file(rev_file_path,
                                                       lock_file=rev_file_path + ".lock")
            rev_list = RevisionList.loads(rev_file)
        else:
            rev_list = RevisionList()
        assert ref.revision is not None, "Invalid revision for: %s" % repr(ref)
        rev_list.add_revision(ref.revision)
        self._storage_adapter.write_file(rev_file_path, rev_list.dumps(),
                                         lock_file=rev_file_path + ".lock")

    def get_package_revisions_references(self, pref):
        """Returns a RevisionList"""
        assert pref.ref.revision is not None, \
            "BUG: server store needs PREV get_package_revisions_references"
        if pref.revision:
            tmp = RevisionList()
            tmp.add_revision(pref.revision)
            return [PkgReference(pref.ref, pref.package_id, rev.revision, rev.time)
                    for rev in tmp.as_list()]

        tmp = self._package_revisions_file(pref)
        ret = self._get_revisions_list(tmp).as_list()
        if not ret:
            raise PackageNotFoundException(pref)
        return [PkgReference(pref.ref, pref.package_id, rev.revision, rev.time) for rev in ret]

    def _get_revisions_list(self, rev_file_path):
        if self._storage_adapter.path_exists(rev_file_path):
            rev_file = self._storage_adapter.read_file(rev_file_path,
                                                       lock_file=rev_file_path + ".lock")
            rev_list = RevisionList.loads(rev_file)
            return rev_list
        else:
            return RevisionList()

    def _get_latest_revision(self, rev_file_path):
        rev_list = self._get_revisions_list(rev_file_path)
        if not rev_list:
            return None
        return rev_list.latest_revision()

    def _recipe_revisions_file(self, ref):
        recipe_folder = normpath(join(self._store_folder, ref_dir_repr(ref)))
        return join(recipe_folder, REVISIONS_FILE)

    def _package_revisions_file(self, pref):
        tmp = normpath(join(self._store_folder, ref_dir_repr(pref.ref)))
        revision = {None: ""}.get(pref.ref.revision, pref.ref.revision)
        p_folder = join(tmp, revision, SERVER_PACKAGES_FOLDER, pref.package_id)
        return join(p_folder, REVISIONS_FILE)

    def get_revision_time(self, ref):
        try:
            rev_list = self._load_revision_list(ref)
        except IOError:
            return None
        return rev_list.get_time(ref.revision)

    def get_package_revision_time(self, pref):
        try:
            rev_list = self._load_package_revision_list(pref)
        except (IOError, OSError):
            return None

        return rev_list.get_time(pref.revision)

    def _remove_revision_from_index(self, ref):
        rev_list = self._load_revision_list(ref)
        rev_list.remove_revision(ref.revision)
        self._save_revision_list(rev_list, ref)

    def _remove_package_revision_from_index(self, pref):
        rev_list = self._load_package_revision_list(pref)
        rev_list.remove_revision(pref.revision)
        self._save_package_revision_list(rev_list, pref)

    def _load_revision_list(self, ref):
        path = self._recipe_revisions_file(ref)
        rev_file = self._storage_adapter.read_file(path, lock_file=path + ".lock")
        return RevisionList.loads(rev_file)

    def _save_revision_list(self, rev_list, ref):
        path = self._recipe_revisions_file(ref)
        self._storage_adapter.write_file(path, rev_list.dumps(), lock_file=path + ".lock")

    def _save_package_revision_list(self, rev_list, pref):
        path = self._package_revisions_file(pref)
        self._storage_adapter.write_file(path, rev_list.dumps(), lock_file=path + ".lock")

    def _load_package_revision_list(self, pref):
        path = self._package_revisions_file(pref)
        rev_file = self._storage_adapter.read_file(path, lock_file=path + ".lock")
        return RevisionList.loads(rev_file)
